// Morseread.c
// v1.1.1

/* Copyright (c) 2003, Rob Bloom
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the <organization> nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY ROB BLOOM ''AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL ROB BLOOM BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdio.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <linux/kd.h>
#include <termios.h>
#include <sys/time.h>
#include <time.h>
#include <signal.h>

#include "morse.h"

#define TONEFREQ 1400	// Default tone freq.  Can be change by command line
			// or config file.

#define DOWN 0
#define UP 1
#define TRANSLATE 2

int c;

struct timeval since;		/* How long has the keyboard been that way? */
char updown =1;			/* Is it up or down now? */

struct timeval tmp;

double sincedub;		/* How long, etc. (in seconds) */
double countlen;		/* How long it takes you to make a count */
int freq;			/* Tone frequency */

double timeelapsed (struct timeval then);

void savelocation();
void gotolocation(int x, int y);
void restorelocation();
void clearline();

void morse (int where, double howlong, double onecount);

void signalled(int signum);
void cleanup();

int eger;			/* ha ha ha */

int first = 1;			/* Have they inputted anything yet?. This is
used in formatting to prevent blank spaces from getting in front of whatever
you type.  Not really necessary, but otherwise it's a bit annoying, IMO. */

static struct termios orig, new;

void beep(int status);

#define ON 1
#define OFF 0

#define lines 60

struct option *opts; // duration and frequency.

// int lines; // Number of lines in display

int main(int argc, char *argv[]) {
  freq = TONEFREQ;
  countlen = -1;
  // First off, before we mess with the TTY, let's parse the arguments. If
  // there's a problem, we can easily quit the program without having to do
  // any cleanup and the like.
  opts = malloc(sizeof(struct option)*4);
  opts[0].name = "len"; // length
  opts[1].name = "freq"; // frequency
  opts[2].name = "help"; // help
  opts[3].name = 0;
  opts[0].has_arg = required_argument;
  opts[1].has_arg = required_argument;
  opts[2].has_arg = no_argument;
  opts[3].has_arg = 0;
  opts[0].val = 'l';
  opts[1].val = 'f';
  opts[2].val = 'h';
  opts[3].val = 0;
  opts[0].flag = opts[1].flag = opts[2].flag = opts[3].flag = 0; // Whee.
  
  while ((eger = getopt_long(argc,argv,"l:f:hH",opts, 0)) != -1 ) {
    switch (eger) {
    case 'l':
      countlen = strtod(optarg,0);
      if (countlen <= 0) {
	printf("Invalid argument for length!\n");
	exit(1);
      }
      break;
    case 'f':
      freq = atoi(optarg);
      if (freq < 0) { // We allow zero, for no beeps.
	printf("Invalid argument for frequency!\n");
	exit(2);
      }
      break;
    case 'h':
    case 'H':
    case '?':
      printf("Morseread\n%s\nby Inuchance <inu@inuchance.net>\n",VER);
      printf("IRcorp Enterprises\nhttp://inuchance.net/\n\n");
      // Now for some actual help...
      printf("Accepted command line arguments:\n");
      printf("  -l, --len         Specify length of short tone, in seconds.\n");
      printf("  -f, --freq        Specify the frequency of the tone used when typing morse.\n");
      printf("  -h, --help        This help.\n\n");
      printf("More on counts, used for morse recognition, can be found in the README.\n");
      printf("Also, if you specify the length of the short tone here, you won't need to\n");
      printf("do the short tone calibration when the program starts up.\n");
      exit(0); // YUO AER HALPED
      break;
    }
  }
  atexit(cleanup);
  for (eger = 1; eger <= 32; eger++)
    if (eger != 17)  // Enter generates that, let's ignore it.
    signal(eger,signalled); 	/* It is vital we restore the TTY */
  tcgetattr (0, &orig);
  memcpy(&new,&orig,sizeof(struct termios));
  new.c_lflag &=~ (ICANON | ECHO | ISIG);
  new.c_cc[VMIN] = 0;
  new.c_cc[VTIME] = 1;
  tcsetattr (0, TCSANOW, &new);
  printf("%c[2J",27);
  printf("%c[6;%ir",27,lines); // Set scrolling region.
  gotolocation(1,3);
  printf("Use space bar to type in morse code. Hit escape when you are finished.\nAlso, to translate a line of morse code into plain text, hit enter.");
  ioctl(0, KDSKBMODE, K_MEDIUMRAW);
  gettimeofday(&since,0);
  
  if (countlen == -1) {
    gotolocation(1,1);
    printf("Please make a short tone, for speed calibration.\n");
  }
  gotolocation(1,5);
  while (c != 1) {
    if (read(0, &c, 1)) {
      if ( c == 28 ) { // Enter pushed.
	printf("\n");  // Give a newline to make things look pretty.
	morse(TRANSLATE,1,1);
	first = 1;
      }
      if ( (c & 127) == 57) { // Space bar status modified somehow.
	if ( c & 128 ) { // Space bar up
	  beep(OFF);
	  sincedub = timeelapsed(since);
	  savelocation(); gotolocation(1,1); clearline();
	  printf("\rSpace DOWN for %f seconds.                  \n",sincedub);
	  restorelocation();
	  if (countlen == -1) {
	    countlen = sincedub;
	    printf("Calibrated. One count will be %f seconds long.\n",
		   sincedub);
	    beep(OFF);
	    first=1;
	  } else {
	    morse(DOWN,sincedub,countlen);
	  }
	  gettimeofday(&since, 0);
	  updown = UP;
	    
	} else { // Space bar down
	  beep(ON);
	  if (updown == UP) {	/* Is this the first time? */
	    updown = DOWN;
	    sincedub = timeelapsed(since);
	    savelocation(); gotolocation(1,1); clearline();
	    printf("\r  Space UP for %f seconds.                  \n",sincedub);
	    restorelocation();
	    morse(UP,sincedub,countlen);
	    gettimeofday(&since, 0);
	  }
	  /* Ignore multiple space downs which keyboards like to make */
	}
      }
    }
    savelocation(); gotolocation(1,2);
    printf("\r... %f", timeelapsed(since));
    restorelocation();
    fflush(stdout);
  }
  exit(0);
}

double timeelapsed (struct timeval then) {
  struct timeval now;
  double duration;
  gettimeofday(&now,0);
  
  duration = (double)(now.tv_sec +
		     (double)(now.tv_usec)/1000000);
  duration -= (double)(then.tv_sec +
		      (double)(then.tv_usec)/1000000);
  return duration;
}

void savelocation() {
  printf("%c[s",27);
  return;
}

void gotolocation(int x, int y) {
  printf("%c[%i;%iH",27,y,x);
  return;
}

void restorelocation() {
  printf("%c[u",27);
  return;
}

void clearline() {
  printf("%c[[2K",27);
  return;
}

void morse (int where, double howlong, double onecount) {
  double counts;
  counts = howlong/onecount;
  static char *morseline = 0;
  if (!morseline)
    morseline = malloc(16384); // You'd have to be crazy to morse that much
                               // code, I think.
  switch (where) {
  case DOWN:
    //    beep(OFF);
    if (counts > 0 && counts < 1.5) {
      putchar('.');
      sprintf(morseline+strlen(morseline),".");
    } else if (counts >= 1.5) {
      putchar('-');
      sprintf(morseline+strlen(morseline),"-");
    }
    break;
  case UP:
    //    beep(ON);
    if (!first) {
      if (counts > 0 && counts < 1.5) {
	// Inter-morse char. (dit daw)
      } else if (counts >= 1.5 && counts < 4.5) {
	// Inter-letter char. (a b)
	printf(" ");
	sprintf(morseline+strlen(morseline)," ");
      } else if (counts >= 4.5) {
	// Inter-word char. (foo bar)
	printf("  ");
	// Old /usr/games/morse compatible method.
	// sprintf(morseline+strlen(morseline),"\n,\n");
	// Shiny new method designed for my internal function.
	sprintf(morseline+strlen(morseline)," , ");
      }
    }
    first = 0;
    break;
  case TRANSLATE:
/*  Actually, I think I'll implement this myself... More compatible that way.
    // Call /usr/games/morse -d
    link = popen("/usr/games/morse -d", "w");
    fprintf(link,morseline);
    fprintf(link,"\n");
    pclose(link);
*/
    printf("%s\n",translate_morse(morseline));
    morseline[0] = 0; // Clear the string.
  }
}

void signalled(int signum) {
  // Doesn't matter what signal we get, but signal() is picky.
  cleanup();
  printf("Received signal %i, exiting.\n",signum);
  exit(signum);
}

void cleanup() {
  ioctl(0, KDSKBMODE, K_XLATE);
  printf("%c[1;%ir",27,lines); //Repair scrolling region
  tcsetattr(0, TCSANOW, &orig);
  printf ("%c[u",27); // Restore last saved position.
  beep(OFF);
  printf("\n");
}

void beep (int status) {
  static int isbeeping = 0; // On some stupider machines, (such as my 486
                            // upstairs) beeps get choppy from multiple issued
                            // beep commands. Thusly, this is to nullify that.
  if (freq != 0) {
    switch (status) {
    case OFF:
      if (isbeeping) {
	ioctl(0,KIOCSOUND,0);
	isbeeping = 0;
      }
      break;
    case ON:
      if (!isbeeping) {
	ioctl(0,KIOCSOUND, 1190000/freq);
	isbeeping=1;
      }
      break;
    }
  }
}

